%{

/* 
   def.l -- PIKT piktc def lexer:  deals with define conditionals and the
            marking of config file lines for processing; marking does not
            apply to systems.cfg!

   $Header: /var/cvs/PIKT/pikt/src/piktc/def.l,v 1.38 2006/11/04 09:59:35 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2006 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org

   ----------------------------------------------------------------------------

   we implement '#setdef foo = TRUE' by throwing back to the lexer the
   faked input '\n#define foo'

   we implement '#setdef foo = FALSE' by throwing back to the lexer the
   faked input '\n#undefine foo'

   we implement '#setdef foo = bar' by throwing back to the lexer the
   faked input '\n#ifdef bar\n#define foo\n#elsedef\n#undefine foo\n#endifdef'

 */

#include "../lib/common.h"
#include "piktc.h"
#include "def.tab.h"

char _deflex_[] = "deflex";

#define RETURN(token) deflval.sval = mystrdup(yytext, _deflex_); return token

#undef YY_INPUT
#define YY_INPUT(buf, result, max_size) { \
		signed char c = definput(); \
		result = (c == EOF) ? YY_NULL : (buf[0] = c, 1); \
	}

cfgln *gclp;	/* global config line pointer, tracking the current
		   config line for purposes of setting its doproc;
		   we could make this a truly global var, replacing
		   "$1->doproc" with "gclp->doproc" in def.y, and
		   thereby eliminating the need for the
		   "deflval.clpval = gclp" below, but we prefer
		   explicit parameter passing in this case (even
		   as it imposes a little additional overhead)
		 */

/* #setdef stuff */
int eqnot;
char *def1;
unsigned int def1len;
char *sd;
unsigned int sdlen;
int ii;

void deferror(const char *);
signed int definput(void);

%}

W		[ \t]
D		[0-9]
L		[A-Za-z]
DL		[0-9A-Za-z]

%x DS0 DS1 DS2 DS3 DS4

%option noyywrap

%%

<*>"/"{2,8}\n			{ switch(strlen(yytext)-1) {
				  	case 2:
						BEGIN(DS0);
						return _DEFCFGBOF;
						break;
				  	case 3:
						BEGIN(INITIAL);
						return _MCRCFGBOF;
						break;
				  	case 4:
						BEGIN(INITIAL);
						return _ALMCFGBOF;
						break;
				  	case 5:
						BEGIN(INITIAL);
						return _ALTCFGBOF;
						break;
				  	case 6:
						BEGIN(INITIAL);
						return _PRGCFGBOF;
						break;
				  	case 7:
						BEGIN(INITIAL);
						return _FILCFGBOF;
						break;
				  	case 8:
						BEGIN(INITIAL);
						return _OBJCFGBOF;
						break;
				  }
				}

<*>^#{W}*(def|define){W}+	{ BEGIN(DS1);
				  return _DEF;
				}
<*>^#{W}*(undef|undefine){W}+	{ BEGIN(DS1);
				  return _UNDEF;
				}
<*>^#{W}*ifdef{W}+		{ BEGIN(DS1);
				  return _IFDEF;
				}
<*>^#{W}*ifndef{W}+		{ BEGIN(DS1);
				  return _IFNDEF;
				}
<*>^#{W}*(elifdef|elsifdef|elseifdef){W}+ {
				  BEGIN(DS1);
				  return _ELIFDEF;
				}
<*>^#{W}*(elifndef|elsifndef|elseifndef){W}+ {
				  BEGIN(DS1);
				  return _ELIFNDEF;
				}
<*>^#{W}*elsedef{W}*\n		{ BEGIN((cfgcurr == DEFCFG) ? DS0 : INITIAL);
				  return _ELSEDEF;
				}
<*>^#{W}*(endifdef|fidef){W}*\n? {
				  BEGIN((cfgcurr == DEFCFG) ? DS0 : INITIAL);
				  return _ENDIFDEF;
				}

<*>^#{W}*setdef{W}+		{ BEGIN(DS2); }

<*>\177				{ ; /* dummy char; see
				       definput() below
				     */
				}

<DS0>^{L}[-_\.0-9a-zA-Z]*	{ RETURN(_STANZA); }
<DS0>1|true|TRUE|yes|YES|on|ON	{ RETURN(_TRUE); }
<DS0>0|false|FALSE|no|NO|off|OFF { RETURN(_FALSE); }
<DS0>\[[^\n]+\]			{ *strrchr(yytext, ']') = '\0';
				  if (system(yytext+1)) {
					RETURN(_FALSE);
				  } else {
					RETURN(_TRUE);
				  }
				}
<DS0>{W}			{ ; }
<DS0>\n				{ ; }

<DS1>{L}[-_\.0-9a-zA-Z]*	{ BEGIN((cfgcurr == DEFCFG) ? DS0 : INITIAL);
				  RETURN(_IDENT);
				}

<DS2>{L}[-_\.0-9a-zA-Z]*	{ def1 = mystrdup(yytext, _deflex_);
				  def1len = yyleng;
				  BEGIN(DS3);
				}
<DS3>{W}*={W}*			{ eqnot = FALSE;
				  BEGIN(DS4);
				}
<DS3>{W}*={W}*\!{W}*		{ eqnot = TRUE;
				  BEGIN(DS4);
				}
<DS4>0|false|FALSE|no|NO|off|OFF|1|true|TRUE|yes|YES|on|ON {
				  sdlen = def1len + 11;
				  /* 11 is length of the sprintf() format
				     string below (2 for the possible "un";
				     we +1 to sdlen when doing the mycalloc()
				     to account for the string-ending '\0'
				   */
				  sd = mycalloc((unsigned int)(sdlen+1), sizeof(char), _deflex_);
				  if (eqnot) {
					  sprintf(sd, "\n#%sdefine %s", (strchr("1tTyY", *yytext) || ! strcasecmp(yytext, "on")) ? "un" : nullstring, def1);
				  } else {
					  sprintf(sd, "\n#%sdefine %s", (strchr("0fFnN", *yytext) || ! strcasecmp(yytext, "off")) ? "un" : nullstring, def1);
				  }
				  ii = sdlen;
				  while (ii > 0) {
					/* ignore possible extra padding
					   at end of string, because sd
					   may or may not have "un"
					 */
					if (*(sd+ii-1) != '\0')
						unput(*(sd+ii-1));
					ii--;
				  }
				  free(def1);
				  free(sd);
				  BEGIN((cfgcurr == DEFCFG) ? DS0 : INITIAL);
				}
<DS4>\[[^\n]+\]			{ *strrchr(yytext, ']') = '\0';
				  sdlen = def1len + 11;
				  /* 11 is length of the sprintf() format
				     string below (2 for the possible "un";
				     we +1 to sdlen when doing the mycalloc()
				     to account for the string-ending '\0'
				   */
				  sd = mycalloc((unsigned int)(sdlen+1), sizeof(char), _deflex_);
				  if (eqnot) {
					  sprintf(sd, "\n#%sdefine %s", ! system(yytext+1) ? "un" : nullstring, def1);
				  } else {
					  sprintf(sd, "\n#%sdefine %s", system(yytext+1) ? "un" : nullstring, def1);
				  }
				  ii = sdlen;
				  while (ii > 0) {
					/* ignore possible extra padding
					   at end of string, because sd
					   may or may not have "un"
					 */
					if (*(sd+ii-1) != '\0')
						unput(*(sd+ii-1));
					ii--;
				  }
				  free(def1);
				  free(sd);
				  BEGIN((cfgcurr == DEFCFG) ? DS0 : INITIAL);
				}
<DS4>{L}[-_\.0-9a-zA-Z]*	{ sdlen = 2*def1len + yyleng + 47;
				  /* 47 is length of the sprintf() format
				     string below less the '%s';
				     we +1 to sdlen when doing the mycalloc()
				     to account for the string-ending '\0'
				   */
				  sd = mycalloc((unsigned int)(sdlen+1), sizeof(char), _deflex_);
				  sprintf(sd, "\n#ifdef %s\n#%sdefine %s\n#elsedef\n#%sdefine %s\n#endifdef", yytext, eqnot ? "un" : nullstring, def1, eqnot ? nullstring : "un", def1);
				  ii = sdlen;
				  while (ii > 0) {
					unput(*(sd+ii-1));
					ii--;
				  }
				  free(def1);
				  free(sd);
				  BEGIN((cfgcurr == DEFCFG) ? DS0 : INITIAL);
				}
<DS3,DS4>{W}			{ ; }

<INITIAL>^[^#\n][^\n]*\n?	{ deflval.clpval = gclp;
				  return _LINE;
				}

<*>!				{ return yytext[0]; }
<*>{W}				{ ; }
<*>\n				{ ; }

<*>.				{ deferror("syntax error"); }

%%

/* def error handler
 */
void
deferror (const char *message)
{

	if (cfgline && strcmp(cfgline,cfgbof[cfgcurr]) && (cfgcurr <= cfglast)) {
		logreturn(OUTPUT, LOG_ERR, ERRID_DEF_PARSE_ERROR, __FILE__, __LINE__, "defparse", FALSE, "%s, at or near line %d: %s: %s", cfgfilecurr, lineno, message, yytext ? yytext : nullstring);
	} else {
		logreturn(OUTPUT, LOG_ERR, ERRID_DEF_PARSE_ERROR, __FILE__, __LINE__, "defparse", FALSE, "%s, at or near line %d: %s", cfgfileprev ? cfgfileprev : cfgfilecurr, lineno, message);
	}
	bailout = TRUE;
	BEGIN((cfgcurr == DEFCFG) ? DS0 : INITIAL);
	return;
}

/* for all config files except systems.cfg, feed their contents to the
   lexer input stream
 */
signed int
definput (void)
{
	static char fnnam[] = "definput";
	static char *line;
	static cfgln *clp;
	register char c;

	if ((cfgcurr > cfglast) /* || (cfgcurr > OBJCFG) */) {
		loffset = 0;
		return EOF;
	}

	if (! loffset) {
		if (bof) {
			cfgline = line = (char *) cfgbof[cfgcurr];
			clp = configs[cfgcurr];
			gclp = clp;
			cfgfileprev = cfgfilecurr;
			cfgfilecurr = (char *) cfgfilenames[cfgcurr];
			/* zero this flag; at end of config lines, if this
			   is still zero, we know that no config lines were
			   processed for current host
			 */
			lineno = 0;
			if (! clp) {
				if (isverboselog)
					logreturn(PIKTLOG, LOG_WARNING, ERRID_CFG_FILE_MISSING, __FILE__, __LINE__, fnnam, FALSE, "missing %s", cfgfilecurr);
				cfgcurr++;
				if ((cfgcurr > cfglast) /* || (cfgcurr > OBJCFG) */) {
					loffset = 0;
					return EOF;
				} else {
					loffset = 0;
					bof = TRUE;
					return '\177';	/* dummy char */
				}
			}
		} else {
			while (! bsmember(clp->bs, hbscurr, hbslongs)) {
				clp = clp->next;
				if (clp) {
					cfgfileprev = cfgfilecurr;
					cfgfilecurr = clp->file;
				} else {
					if (lineno == 0)
						logexit(OUTPUT, LOG_ERR, ERRID_CFG_NO_INPUT, __FILE__, __LINE__, fnnam, FALSE, "%s, for host %s, no input!", cfgfileprev ? cfgfileprev : cfgfilecurr, hostcurr);
					cfgcurr++;
					if ((cfgcurr > cfglast) /* || (cfgcurr > OBJCFG) */) {
						loffset = 0;
						return EOF;
					} else {
						loffset = 0;
						bof = TRUE;
						return '\177';	/* dummy char */
					}
				}
			}
			cfgline = line = clp->line;
			lineno = clp->lineno;
		}
	}

	if ((c = *(line+loffset++)) == '\n') {
		if (bof)
			bof = FALSE;
		else {
			gclp = clp;
			clp = clp->next;
			if (clp) {
				cfgfileprev = cfgfilecurr;
				cfgfilecurr = clp->file;
			} else {
				cfgcurr++;
				bof = TRUE;
			}
		}
		loffset = 0;
	}

	return c;

}
