%{

/*
   cmd.y -- PIKT pikt command-line parser

   $Header: /var/cvs/PIKT/pikt/src/piktd/cmd.y,v 1.29 2006/11/04 09:59:36 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2006 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#include "../lib/common.h"
#include "piktd.h"

char _cmdparse_[] = "cmdparse";

char *cp;

int k;

vl *vlp;

char ac[16];

int MYARGC = 0;

%}

%union {
	signed char *sval;
}

%token _MINUS _PLUS _OPT _CHECK _DOUBLECHECK _MAILCMD _LPCMD _ALERT _SCRIPT _COMMAND
%token <sval> _ITEM

%%

start:		/* empty */ { ishelp = TRUE; }
		| options
		| items
		| options items
		;

options:	option
		| options option
		;

option:		_MINUS opts
		| _PLUS opts /* no plus opts now, maybe in future */
		;

opts:		_OPT
		| opts _OPT
		;

items:		commanditem { interactive = TRUE; }
		| scriptitem { interactive = TRUE; }
		| mail scriptitem { interactive = FALSE; }
		| print scriptitem { interactive = FALSE; }
		| mail print scriptitem { interactive = FALSE; }
		| print mail scriptitem { interactive = FALSE; }
		;

mail:		_PLUS _MAILCMD _ITEM { mailcmd = $3; }
		;

print:		_PLUS _LPCMD _ITEM { lpcmd = $3; }
		;

commanditem:	_PLUS _COMMAND _ITEM {
			isalert = FALSE;
			isscript = TRUE;
			command = $3;
			altcurr = mystrdup(nullstring, _cmdparse_);
			regalm(altcurr);
		}
		;

scriptitem:	alert
		| script
		;

alert:		_PLUS _ALERT _ITEM {
			isalert = TRUE;
			isscript = FALSE;
			altcurr = $3;
		}
		;

script:		_PLUS _SCRIPT _ITEM {
			isalert = FALSE;
			isscript = TRUE;
			altcurr = $3;
			regalm(altcurr);
			initvals(&valnow);
			for (k=0; k<=MAXARYDIM; k++)
				arndx[k] = 0L;
			regvar(mystrdup("ARGV", _cmdparse_), FIXEDVAL, STRING, NUMBER, 1, arndx, (char *)NULL, (char *)NULL);
			if (! (vlp = regvl(valnow, 1, arndx, (char *)NULL, NUMBER))) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_REGVL_ERROR, __FILE__, __LINE__, _cmdparse_, FALSE, "regvl() error for $ARGV[0]");
				bailout = TRUE;
			} else {
				vlp->str = mystrdup($3, _cmdparse_);
			}
			regvar(mystrdup("ARGC", _cmdparse_), FIXEDVAL, NUMBER, NUMBER, 2, arndx, (char *)NULL, (char *)NULL);
			if (! (vlp = regvl(valnow, 2, arndx, (char *)NULL, NUMBER))) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_REGVL_ERROR, __FILE__, __LINE__, _cmdparse_, FALSE, "regvl() error for #ARGC");
				bailout = TRUE;
			} else {
				vlp->str = mystrdup("0", _cmdparse_);
				vlp->num = (double) 0;
			}
		}
		argvs
		| _ITEM {
			isalert = FALSE;
			isscript = TRUE;
			if (strrchr($1, '/')) {
				FREE(prgdir);
				prgdir = mystrdup($1, _cmdparse_);
				cp = strrchr(prgdir, '/');
				*cp = '\0';
				altcurr = cp+1;
			} else {
				altcurr = $1;
			}
			regalm(altcurr);
			initvals(&valnow);
			for (k=0; k<=MAXARYDIM; k++)
				arndx[k] = 0L;
			regvar(mystrdup("ARGV", _cmdparse_), FIXEDVAL, STRING, NUMBER, 1, arndx, (char *)NULL, (char *)NULL);
			if (! (vlp = regvl(valnow, 1, arndx, (char *)NULL, NUMBER))) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_REGVL_ERROR, __FILE__, __LINE__, _cmdparse_, FALSE, "regvl() error for $ARGV[0]");
				bailout = TRUE;
			} else {
				vlp->str = mystrdup($1, _cmdparse_);
			}
			regvar(mystrdup("ARGC", _cmdparse_), FIXEDVAL, NUMBER, NUMBER, 2, arndx, (char *)NULL, (char *)NULL);
			if (! (vlp = regvl(valnow, 2, arndx, (char *)NULL, NUMBER))) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_REGVL_ERROR, __FILE__, __LINE__, _cmdparse_, FALSE, "regvl() error for #ARGC");
				bailout = TRUE;
			} else {
				vlp->str = mystrdup("0", _cmdparse_);
				vlp->num = (double) 0;
			}
		}
		argvs
		;

argvs:		/* empty */
		| argvs argv
		;

argv:		_ITEM {
			arndx[1] = ++MYARGC;
			if (! (vlp = regvl(valnow, 1, arndx, (char *)NULL, NUMBER))) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_REGVL_ERROR, __FILE__, __LINE__, _cmdparse_, FALSE, "regvl() error for ARGV[%d]", MYARGC);
				bailout = TRUE;
			} else {
				vlp->str = $1;
			}
			arndx[1] = 0L;
			if ((vlp = lookupvl(valnow, 2, arndx, (char *)NULL, NUMBER))) {
				FREE(vlp->str);
			} else if (! (vlp = regvl(valnow, 2, arndx, (char *)NULL, NUMBER))) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_REGVL_ERROR, __FILE__, __LINE__, _cmdparse_, FALSE, "regvl() error for #ARGC");
				bailout = TRUE;
			}
			if (vlp) {
				sprintf(ac, "%d", MYARGC);
				vlp->str = mystrdup(ac, _cmdparse_);
				vlp->num = (double) MYARGC;
			}
		}
		;

%%
